@@ -8,17 +8,21 @@
 
 #include "./platforms/WaylandPlatform.hpp"
 #include "../renderer/gl/OpenGL.hpp"
-#include "../window/WaylandWindow.hpp"
+#include "../output/WaylandOutput.hpp"
 #include "../window/WaylandLayer.hpp"
+#include "../window/WaylandLockSurface.hpp"
+#include "../window/WaylandWindow.hpp"
 #include "../Macros.hpp"
 #include "../element/Element.hpp"
 #include "../palette/ConfigManager.hpp"
 #include "../system/Icons.hpp"
+#include "../sessionLock/WaylandSessionLock.hpp"
 
 #include <sys/wait.h>
 #include <sys/poll.h>
 
 #include <print>
+#include <unistd.h>
 
 using namespace Hyprtoolkit;
 using namespace Hyprutils::Memory;
@@ -74,6 +78,13 @@ SP<IBackend> IBackend::create() {
         g_logger->log(HT_LOG_ERROR, "couldn't start aq backend");
         return nullptr;
     }
+    g_waylandPlatform = makeUnique<CWaylandPlatform>();
+    if (!g_waylandPlatform->attempt()) {
+        g_waylandPlatform = nullptr;
+        return nullptr;
+    }
+    g_openGL   = makeShared<COpenGLRenderer>(g_waylandPlatform->m_drmState.fd);
+    g_renderer = g_openGL;
 
     return g_backend;
 };
@@ -90,14 +101,27 @@ SP<CPalette> CBackend::getPalette() {
     return g_palette;
 }
 
+std::vector<SP<IOutput>> CBackend::getOutputs() {
+    if (!g_waylandPlatform)
+        return {};
+
+    return std::vector<SP<IOutput>>(g_waylandPlatform->m_outputs.begin(), g_waylandPlatform->m_outputs.end());
+}
+
+std::expected<SP<ISessionLockState>, eSessionLockError> CBackend::aquireSessionLock() {
+    if (!g_waylandPlatform)
+        return std::unexpected(LOCK_ERROR_PLATFORM_UNINITIALIZED);
+
+    auto lockState = g_waylandPlatform->aquireSessionLock();
+    if (!lockState || lockState->m_denied)
+        return std::unexpected(LOCK_ERROR_DENIED);
+
+    return lockState;
+}
+
 SP<IWindow> CBackend::openWindow(const SWindowCreationData& data) {
-    if (!g_waylandPlatform) {
-        g_waylandPlatform = makeUnique<CWaylandPlatform>();
-        if (!g_waylandPlatform->attempt())
-            return nullptr;
-        g_openGL   = makeShared<COpenGLRenderer>(g_waylandPlatform->m_drmState.fd);
-        g_renderer = g_openGL;
-    }
+    if (!g_waylandPlatform)
+        return nullptr;
 
     if (data.type == HT_WINDOW_LAYER) {
         if (!g_waylandPlatform->m_waylandState.layerShell)
@@ -108,6 +132,20 @@ SP<IWindow> CBackend::openWindow(const SWindowCreationData& data) {
         w->m_rootElement->impl->window = w;
         g_waylandPlatform->m_layers.emplace_back(w);
         return w;
+    } else if (data.type == HT_WINDOW_LOCK_SURFACE) {
+        if (!g_waylandPlatform->m_waylandState.sessionLock) {
+            g_logger->log(HT_LOG_ERROR, "No session lock manager. Does your compositor support it?");
+            return nullptr;
+        }
+
+        if (!g_waylandPlatform->m_sessionLockState || g_waylandPlatform->m_sessionLockState->m_denied || g_waylandPlatform->m_sessionLockState->m_sessionUnlocked)
+            return nullptr;
+
+        auto w                         = makeShared<CWaylandLockSurface>(data);
+        w->m_self                      = w;
+        w->m_rootElement->impl->window = w;
+        g_waylandPlatform->m_sessionLockState->m_lockSurfaces.emplace_back(w);
+        return w;
     }
 
     auto w                         = makeShared<CWaylandWindow>(data);
